#include <math.h>
#include "adaptSimpson.h"

    

double adaptiveSimpsons_helper(double (*f)(double x),
			   double a, double b, double epsilon,                 
			   double S, double fa, double fb, double fc, long bottom);


// SIMPSON: adaptive SIMPSON from Wikipedia (downloaded 2018)
// from https://en.wikipedia.org/wiki/Adaptive_Simpson%27s_method
//
// Adaptive Simpson's Rule

double adaptiveSimpsons(double (*f)(double x),   // ptr to function
			double a, double b,                 // interval [a,b]
			double epsilon,                     // error tolerance
			long maxRecursionDepth)             // recursion cap        
{
  double c  = (a + b)/2.0;
  double h  = b - a;                                                                  
  double fa = f(a), fb = f(b);
  double fc = f(c);                                                           
  double S  = (h/6.0)*(fa + 4.0*fc + fb);                                                                
  return adaptiveSimpsons_helper(f, a, b, epsilon, S, fa, fb, fc, maxRecursionDepth);                   
}                                                                                                   
 
//
// Recursive helper function for adaptiveSimpsons() function
//                                                                                                 
double adaptiveSimpsons_helper(double (*f)(double x),
			   double a, double b, double epsilon,                 
			   double S, double fa, double fb, double fc, long bottom)
{                 
  double c = (a + b)/2;
  double h = b - a;                                                                  
  double d = (a + c)/2;
  double e = (c + b)/2;                                                              
  double fd = f(d);
  double fe = f(e);                                                                      
  double Sleft = (h/12)*(fa + 4*fd + fc);                                                           
  double Sright = (h/12)*(fc + 4*fe + fb);                                                          
  double S2 = Sleft + Sright;                                                                       
  if (bottom <= 0 || fabs(S2 - S) <= 15.0*epsilon)   // magic 15 comes from error analysis                                       
    return S2 + (S2 - S)/15.0;                                                                        
  return adaptiveSimpsons_helper(f, a, c, epsilon/2.0, Sleft,  fa, fc, fd, bottom-1) +                    
    adaptiveSimpsons_helper(f,  c, b, epsilon/2.0, Sright, fc, fb, fe, bottom-1);                     
}
